#!/usr/bin/env python3
# -----------------------------------------------------------------------------
# Programmer(s): David J. Gardner @ LLNL
# -----------------------------------------------------------------------------
# SUNDIALS Copyright Start
# Copyright (c) 2002-2024, Lawrence Livermore National Security
# and Southern Methodist University.
# All rights reserved.
#
# See the top-level LICENSE and NOTICE files for details.
#
# SPDX-License-Identifier: BSD-3-Clause
# SUNDIALS Copyright End
# -----------------------------------------------------------------------------
# Script to plot and compare time series data from multiple files
#
# The input data files must contain an M x (N + 1) matrix of values where the
# first column of each row is the output time and the remaining values are the
# N output quantities. Additionally the data files must contain header comment
# starting with "vars:" that lists the name for each data column i.e.,
#
# # vars: time quantity_0 quantity_1 . . . quantity_N-1
# t_0   q_0,0   q_0,1   q_0,2   . . . q_0,N-1
# t_1   q_1,0   q_1,1   q_1,2   . . . q_1,N-1
#   .     .       .       .       .     .
#   .     .       .       .       .     .
#   .     .       .       .       .     .
# t_M-1 q_M-1,0 q_M-1,1 q_M-1,2 . . . q_M-1,N-1
#
# where t_m is the m-th output time and q_m,n is the n-th quantity at the m-th
# output time.
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# main routine
# -----------------------------------------------------------------------------
def main():

    import argparse
    import matplotlib.pyplot as plt
    import numpy as np
    import shlex

    parser = argparse.ArgumentParser(description='''Plot data files''')

    parser.add_argument('quantity', type=str,
                        help='''Quantity to plot''')

    parser.add_argument('datafiles', type=str, nargs='+',
                        help='''Data files to plot''')

    # Plot display options

    parser.add_argument('--save', action='store_true',
                        help='''Save figure to file''')

    parser.add_argument('--labels', type=str, nargs='+',
                        help='''Data file labels for plot legend''')

    parser.add_argument('--title', type=str,
                        help='''Plot title''')

    parser.add_argument('--xlabel', type=str,
                        help='''x-axis label''')

    parser.add_argument('--ylabel', type=str,
                        help='''y-axis label''')

    parser.add_argument('--grid', action='store_true',
                        help='''Add grid to plot''')

    # Axis scaling
    logscale = parser.add_mutually_exclusive_group()

    logscale.add_argument('--logx', action='store_true',
                          help='''Plot with log scale x-axis''')

    logscale.add_argument('--logy', action='store_true',
                          help='''Plot with log scale y-axis''')

    logscale.add_argument('--loglog', action='store_true',
                          help='''Use log scale x and y axes''')

    # Debugging options

    parser.add_argument('--debug', action='store_true',
                        help='Enable debugging')

    # Parse command line args
    args = parser.parse_args()

    # Create figure and axes
    fig, ax = plt.subplots()

    for i, datafile in enumerate(args.datafiles):

        quantities = None
        with open(datafile) as fn:
            # read the file line by line
            for line in fn:
                # skip empty lines
                if not line.strip():
                    continue
                # exit after reading initial comment lines
                if "#" not in line:
                    break
                # split line into a list
                text = shlex.split(line)
                # extract quantity names
                if "vars:" in line:
                    quantities = text[2:]
                    continue

        # Check inputs
        if quantities is None:
            print("ERROR: quantity names not provided")
            sys.exit()

        if args.quantity not in quantities:
            print("ERROR: quantity not found")
            print(f"  Possible values: {quantities}")
            sys.exit()

        # Get data column index
        idx = quantities.index(args.quantity)

        # Load data
        data = np.loadtxt(datafile, dtype=np.double)

        if args.debug:
            print(np.shape(data))
            print(data)

        # Extract t and q data
        tdata = data[:,0]    # first column has t values
        qdata = data[:,idx]  # remaining columns have q values

        # line colors: matplotlib.org/stable/tutorials/colors/colormaps.html
        # and colorbrewer2.org)
        if len(args.datafiles) < 22:
            colors = ["#d62728", "#1f77b4", "#2ca02c", "#9467bd", "#ff7f0e",
                      "#8c564b", "#e377c2", "#7f7f7f", "#bcbd22", "#17becf",
                      "#000000", "#ff9896", "#aec7e8", "#98df8a", "#c5b0d5",
                      "#ffbb78", "#c49c94", "#f7b6d2", "#c7c7c7", "#dbdb8d",
                      "#9edae5"]
        else:
            print("ERROR: ncols > ncolors")
            sys.exit()

        # Set plot label for legend
        if (args.labels):
            label=args.labels[i]
        else:
            label=None

        if args.logx or args.logy or args.loglog:
            ax.plot(tdata, np.abs(qdata), label=label,
                    color=colors[i])
        else:
            ax.plot(tdata, qdata, label=label,
                    color=colors[i])

    # Change axis scale
    if args.logx:
        ax.set_xscale("log")
    elif args.logy:
        ax.set_yscale("log")
    elif args.loglog:
        ax.set_xscale("log")
        ax.set_yscale("log")

    # Add title
    if args.title:
        plt.title(args.title)

    # Add x-axis label
    if args.xlabel:
        plt.xlabel(args.xlabel)
    else:
        plt.xlabel("time")

    # Add y-axis label
    if args.ylabel:
        plt.ylabel(args.ylabel)
    else:
        plt.ylabel(args.quantity.replace("_"," "));

    # Add legend
    if args.labels:
        ax.legend(bbox_to_anchor=(1, 0.95), loc="upper right")

    # Add grid
    if args.grid:
        plt.grid()

    # Save plot to file
    if args.save:
        plt.savefig("fig.pdf")
    else:
        plt.show()

# -----------------------------------------------------------------------------
# run the main routine
# -----------------------------------------------------------------------------

if __name__ == '__main__':
    import sys
    sys.exit(main())



#!/usr/bin/env python3
# ----------------------------------------------------------------
# Programmer(s): Cody J. Balos @ LLNL
# ----------------------------------------------------------------
# SUNDIALS Copyright Start
# Copyright (c) 2002-2024, Lawrence Livermore National Security
# and Southern Methodist University.
# All rights reserved.
#
# See the top-level LICENSE and NOTICE files for details.
#
# SPDX-License-Identifier: BSD-3-Clause
# SUNDIALS Copyright End
# ----------------------------------------------------------------
# matplotlib-based plotting script for ark_kepler.c
# ----------------------------------------------------------------

import numpy as np
import matplotlib.pyplot as plt
import argparse

parser = argparse.ArgumentParser(description='Script for plotting the energy, angular momentum, and phase space solution for ark_kepler.c')
parser.add_argument('output_times', help='file with the output times')
parser.add_argument('solution',  help='file with the solution')
parser.add_argument('conserved_quantities', help='file with conserved quantities')

args = parser.parse_args()

t = np.loadtxt(args.output_times, dtype=np.float64)
y = np.loadtxt(args.solution, dtype=np.float64)
y = np.reshape(y, (y.shape[0]//4, 4))

plt.figure(dpi=200)
plt.plot(y[:,0], y[:,1])
plt.savefig('ark_kepler_phase.png')
plt.close()

conserved = np.loadtxt(args.conserved_quantities, delimiter=',', dtype=np.float64)
energy = conserved[:,0]
energy_0 = conserved[0,0]
L = conserved[:,1]
L_0 = conserved[0,1]

plt.figure(dpi=200)
plt.title('Energy')
plt.plot(t, np.abs(energy))
plt.ylabel('H(t,p,q)')
plt.xlabel('<---  t  --->')
plt.xscale('log')
plt.savefig('ark_kepler_energy.png')
plt.close()

plt.figure(dpi=200)
plt.title('Momentum')
plt.plot(t, L)
plt.ylabel('L(t,p,q)')
plt.xlabel('<---  t  --->')
plt.xscale('log')
plt.savefig('ark_kepler_momentum.png')
plt.close()

#
#  Time plot.
#
plt.figure(dpi=200)
plt.plot(t, y[:,0], linewidth = 2)
plt.plot(t, y[:,1], linewidth = 2)
plt.plot(t, y[:,2], linewidth = 2)
plt.plot(t, y[:,3], linewidth = 2)
plt.grid(True)
plt.legend(['P', 'P\'', 'Q', 'Q\''])
plt.xlabel('<---  t  --->')
plt.ylabel('<---  y(1:4)  --->')
plt.title('Solution in Time')
plt.savefig('ark_kepler_plot.png')
plt.close()

#
#  Phase plot.
#
plt.figure(dpi=200)
plt.plot(y[:,0], y[:,1], linewidth=0.1)
plt.grid(True)
plt.xlabel('<---  y1  --->')
plt.ylabel('<---  y2  --->')
plt.title('Phase Plot')
plt.savefig('ark_kepler_phase.png')
plt.close()



#!/usr/bin/env python3
# -----------------------------------------------------------------------------
# Programmer(s): David J. Gardner @ LLNL
# -----------------------------------------------------------------------------
# SUNDIALS Copyright Start
# Copyright (c) 2002-2024, Lawrence Livermore National Security
# and Southern Methodist University.
# All rights reserved.
#
# See the top-level LICENSE and NOTICE files for details.
#
# SPDX-License-Identifier: BSD-3-Clause
# SUNDIALS Copyright End
# -----------------------------------------------------------------------------
# Script to update example output files for failed tests
#
# Example usage:
#   $ ./updateOutFiles.py ../build ../.
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# main routine
# -----------------------------------------------------------------------------
def main():

    import argparse
    import sys, os, shutil

    parser = argparse.ArgumentParser(description='Update output files')

    parser.add_argument('source', type=str,
                        help='Full path of build directory to read files from')
    parser.add_argument('destination', type=str,
                        help='Full path of sundials location to write files to')
    parser.add_argument('--all','-a', action='store_true',
                        help='Update all output files')
    parser.add_argument('--copy','-c', action='store_true',
                        help='Copy file to destination if not found')
    parser.add_argument('--verbose','-v', action='count', default=0,
                        help='Enable verbose output')


    # parse command line args
    args = parser.parse_args()

    # check inputs
    if (not os.path.isdir(args.source)):
        print(f"Error: could not find {args.source}")
        return -1

    if (not os.path.isdir(args.destination)):
        print(f"Error: could not find {args.destination}")
        return -1

    # check that the output directory exists
    output = os.path.join(args.source, "Testing", "output")
    if (not os.path.isdir(output)):
        print(f"Error: could not find {output}")
        return -1

    # create a list of all test run or just the failed tests
    tests = []

    if (args.all):
        # get names of all .out files
        for f in os.listdir(output):
            if f.endswith(".out"):
                tests.append(f)
    else:
        failed = os.path.join(args.source, "Testing", "Temporary", "LastTestsFailed.log")

        # extract test names from list and append .out
        with open(failed, 'r') as f:
            for line in f:
                tests.append(line.split(':')[1].rstrip() + ".out")

    # if failed tests were found update the output files
    if tests:
        paths = [os.path.join(args.destination, "examples"),
                 os.path.join(args.destination, "test", "unit_tests"),
                 args.destination]
        for t in tests:
            if (args.verbose > 0):
                print(f"Searching for {t}")
            found = False
            for p in paths:
                if (args.verbose == 2):
                    print(f"  Looking in {p}")
                for root, dirs, files in os.walk(p):
                    if (args.verbose == 3):
                        print(f"  Looking in {root}")
                    if t in files:
                        if (args.verbose == 1):
                            print(f"  Found file in {root}")
                        if (args.verbose > 1):
                            print("  Found file")
                        shutil.copy(os.path.join(output, t), os.path.join(root, t))
                        found = True
                        break
                if found:
                    break
            if not found:
                if args.copy:
                    print(f"Warning: did not find {t}, copying to {args.destination}")
                    shutil.copy(os.path.join(output, t), os.path.join(args.destination, t))
                else:
                    print(f"Warning: did not find {t}")

# -----------------------------------------------------------------------------
# run the main routine
# -----------------------------------------------------------------------------
if __name__ == '__main__':
    import sys
    sys.exit(main())



# ----------------------------------------------------------------
# SUNDIALS Copyright Start
# Copyright (c) 2002-2024, Lawrence Livermore National Security
# and Southern Methodist University.
# All rights reserved.
#
# See the top-level LICENSE and NOTICE files for details.
#
# SPDX-License-Identifier: BSD-3-Clause
# SUNDIALS Copyright End
# ----------------------------------------------------------------
doc_version = 'v7.1.1'
sundials_version = 'v7.1.1'
arkode_version = 'v6.1.1'
cvode_version = 'v7.1.1'
cvodes_version = 'v7.1.1'
ida_version = 'v7.1.1'
idas_version = 'v6.1.1'
kinsol_version = 'v7.1.1'
year = '2024'

# Warn about all references where the target cannot be found
nitpicky = True

# List of tuples (type, target) to ignore when generating "nitpicky" warnings
nitpick_ignore = [
    # C/C++ types Sphinx does not seem to know about
    ('c:identifier', 'FILE'),
    ('cpp:identifier', 'FILE'),
    ('c:identifier', 'size_t'),
    ('cpp:identifier', 'size_t'),
    # CUDA
    ('cpp:identifier', 'cudaStream_t'),
    ('c:identifier', 'cusparseHandle_t'),
    ('c:identifier', 'cusparseMatDescr_t'),
    ('c:identifier', 'cusolverHandle_t'),
    # Ginkgo
    ('cpp:identifier', 'gko'),
    ('cpp:identifier', 'gko::dim<2>'),
    ('cpp:identifier', 'gko::Executor'),
    ('cpp:identifier', 'gko::LinOp'),
    # HIP
    ('cpp:identifier', 'hipStream_t'),
    # hypre
    ('c:identifier', 'hypre_ParVector'),
    # Kokkos
    ('cpp:identifier', 'ExecutionSpace'),
    ('cpp:identifier', 'ExecutionSpace::memory_space'),
    ('cpp:identifier', 'Kokkos'),
    ('cpp:identifier', 'Kokkos::DefaultExecutionSpace'),
    ('cpp:identifier', 'Kokkos::RangePolicy<exec_space>'),
    ('cpp:identifier', 'Kokkos::View<sunrealtype***, memory_space>'),
    ('cpp:identifier', 'Kokkos::View<sunrealtype*, MemorySpace>'),
    ('cpp:identifier', 'Kokkos::Rank<3>'),
    ('cpp:identifier', 'Kokkos::TeamPolicy<exec_space>'),
    ('cpp:identifier', 'Kokkos::TeamPolicy<exec_space>::member_type'),
    ('cpp:identifier', 'Kokkos::MDRangePolicy<exec_space, Kokkos::Rank<3>>'),
    # MPI
    ('c:identifier', 'MPI_Comm'),
    # PETSc
    ('c:identifier', 'SNES'),
    ('c:identifier', 'PetscErrorCode'),
    ('c:identifier', 'Vec'),
    # SuperLU
    ('c:identifier', 'gridinfo_t'),
    ('c:identifier', 'SuperMatrix'),
    ('c:identifier', 'gridinfo_t'),
    ('c:identifier', 'xLUstruct_t'),
    ('c:identifier', 'xScalePermstruct_t'),
    ('c:identifier', 'xSOLVEstruct_t'),
    ('c:identifier', 'SuperLUStat_t'),
    ('c:identifier', 'superlu_dist_options_t'),
    # SYCL
    ('cpp:identifier', 'sycl'),
    ('cpp:identifier', 'sycl::queue'),
    # Trilinos
    ('cpp:identifier', 'Tpetra'),
    ('cpp:identifier', 'Tpetra::Vector<sunrealtype, int, sunindextype>'),
    ('cpp:identifier', 'Teuchos'),
    ('cpp:identifier', 'Teuchos::RCP<vector_type>'),
    # XBraid
    ('c:identifier', 'braid_AccessStatus'),
    ('c:identifier', 'braid_App'),
    ('c:identifier', 'braid_BufferStatus'),
    ('c:identifier', 'braid_Core'),
    ('c:identifier', 'braid_Int'),
    ('c:identifier', 'braid_PtFcnAccess'),
    ('c:identifier', 'braid_PtFcnInit'),
    ('c:identifier', 'braid_PtFcnSpatialNorm'),
    ('c:identifier', 'braid_PtFcnStep'),
    ('c:identifier', 'braid_Real'),
    ('c:identifier', 'braid_StepStatus'),
    ('c:identifier', 'braid_Vector'),
    # C types referenced in C++ functions, not sure how to fix
    ('cpp:identifier', 'sunbooleantype'),
    ('cpp:identifier', 'sunindextype'),
    ('cpp:identifier', 'sunrealtype'),
    ('cpp:identifier', 'SUNErrCode'),
    ('cpp:identifier', 'SUNContext'),
    ('cpp:identifier', 'N_Vector'),
    ('cpp:identifier', 'SUNMatrix'),
    ('cpp:identifier', 'SUNLinearSolver'),
    ('cpp:identifier', 'SUNMemoryHelper'),
    ('cpp:identifier', 'SUNMemoryType'),
    # C++ namespaces don't seem to work as expected, not sure how to fix
    ('cpp:identifier', 'sundials'),
    ('cpp:identifier', 'sundials::cuda'),
    ('cpp:identifier', 'sundials::hip'),
    ('cpp:identifier', 'sundials::sycl'),
    ('cpp:identifier', 'sundials::ginkgo'),
    # Experimental or internal namespaces and types
    ('cpp:identifier', 'sundials::impl'),
    ('cpp:identifier', 'sundials::impl::BaseNVector'),
    ('cpp:identifier', 'sundials::impl::BaseMatrix'),
    ('cpp:identifier', 'sundials::impl::BaseLinearSolver'),
    ('cpp:identifier', 'sundials::ConvertibleTo<N_Vector>'),
    ('cpp:identifier', 'sundials::ConvertibleTo<SUNMatrix>'),
    ('cpp:identifier', 'sundials::ConvertibleTo<SUNLinearSolver>'),
    # Defined types in Kokkos vector that are not found by Sphinx
    ('cpp:identifier', 'view_type::size_type'),
    ('cpp:identifier', 'view_type::HostMirror'),
    # Template parameter that causes an error in Kokkos matrix
    ('cpp:identifier', 'MatrixType'),
    # C++ types referenced in "C" functions, should probably switch
    # documentation to use .. cpp:function rather than .. c:function
    ('c:identifier', 'SUNCudaExecPolicy'),
    ('c:identifier', 'SUNHipExecPolicy')
]



#!/usr/bin/env python3
# -----------------------------------------------------------------------------
# Programmer(s): Yu Pan @ LLNL
# -----------------------------------------------------------------------------
# SUNDIALS Copyright Start
# Copyright (c) 2002-2024, Lawrence Livermore National Security
# and Southern Methodist University.
# All rights reserved.
#
# See the top-level LICENSE and NOTICE files for details.
#
# SPDX-License-Identifier: BSD-3-Clause
# SUNDIALS Copyright End
# -----------------------------------------------------------------------------

import os
import glob
import argparse
import multiprocessing as mp

import thicket as tt


def main():
    parser = argparse.ArgumentParser(description='Compare Sundials performance results against previous results')

    parser.add_argument('--release', dest='release', action='store_true', help='indicate if the current run to process is a release')

    parser.add_argument('--calidir', dest='caliDir', type=str, help='path to directory containing caliper files', default="/usr/workspace/sundials/califiles")

    parser.add_argument('--releasedir', dest='releaseDir', type=str, help='path to directory containing release caliper files', default="/usr/workspace/sundials/califiles/Release")

    parser.add_argument('--outpath', dest='outPath', type=str, help='path to directory to write results to', default="/dev/null")

    parser.add_argument('--jobid', dest='jobID', type=int, help='job id of the current run to identify .cali files')

    parser.add_argument('--threshold', dest="threshold", type=float, help='the percentage threshold in performance difference that indicates a regression', default=2.0)

    args = parser.parse_args()

    release = args.release
    releaseDir = args.releaseDir
    caliDir = args.caliDir
    outPath = args.outPath
    jobID = args.jobID
    threshold = args.threshold

    # Get available benchmarks
    benchFiles = glob.glob("%s/Benchmarking/*/*" % caliDir)

    if not os.path.exists(outPath):
        os.makedirs(outPath)
    outFile = open("%s/benchmark_output.out" % outPath, 'w')

    # thread per file
    with mp.Pool() as pool:
        for res in pool.starmap(process_benchmark, [(jobID, release, releaseDir, i, threshold) for i in benchFiles]):
            if res:
                outFile.write(res + "\n")
    outFile.close()

    outFile = open("%s/benchmark_output.out" % outPath, 'r')
    try:
        outLines = outFile.readlines()
    finally:
        outFile.close()

    if (len(outLines) == 0):
        return -1
    return 0

def process_benchmark(jobID, isRelease, releaseDir, benchmarkDir, threshold):
    # Get the current benchmark run
    benchmarkFiles = glob.glob("%s/*.cali" % benchmarkDir)
    # Don't compare if the run didn't include this benchmark
    if (len(benchmarkFiles) == 0):
        return

    th_files = tt.Thicket.from_caliperreader(benchmarkFiles)
    curFilter = lambda x: x['job_id'] == jobID
    th_current = th_files.filter_metadata(curFilter)

    # Get the release caliper file
    cluster = th_current.metadata['cluster'].values[0]
    if isRelease:
        # Get the last release
        versionDirs = glob.glob("%s/%s/*" % (releaseDir, cluster))
        versionDirs.sort(key=os.path.getmtime, reverse=True)
        versionDir = versionDirs[1]
    else:
        # Get the release the run is a part of
        version = th_current.metadata['sundials_version'].values[0]
        versionDir = "%s/%s/%s" % (releaseDir, cluster, version)
    benchmarkName = th_current.metadata['env.TEST_NAME'].values[0]
    releaseFile = glob.glob("%s/Benchmarking/*/%s/*.cali" % (versionDir, benchmarkName), recursive=True)
    th_compare = tt.Thicket.from_caliperreader(releaseFile)
    metrics = ['Max time/rank']
    tt.mean(th_current, columns=metrics)
    tt.mean(th_compare, columns=metrics)

    ratio = th_current.statsframe.dataframe['Max time/rank_mean'] / th_compare.statsframe.dataframe['Max time/rank_mean']

    tolerance = threshold/100
    if 1 - ratio[0] < tolerance:
        return benchmarkName


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -----------------------------------------------------------------------------
# Programmer(s): Cody Balos @ LLNL
# -----------------------------------------------------------------------------
# SUNDIALS Copyright Start
# Copyright (c) 2002-2024, Lawrence Livermore National Security
# and Southern Methodist University.
# All rights reserved.
#
# See the top-level LICENSE and NOTICE files for details.
#
# SPDX-License-Identifier: BSD-3-Clause
# SUNDIALS Copyright End
# -----------------------------------------------------------------------------
# Module of Python functions that may be useful to SUNDIALS developers writing
# scripts to parse logs produced by SUNLogger.
# -----------------------------------------------------------------------------

import re
import numpy as np


def parse_logfile_payload(payload, line_number, all_lines, array_indicator="(:)"):
    """
    This function parses the payload of in a SUNDIALS log file line
    into a dictionary. The payload of a SUNDIALS log file line
    is the part after all the [ ] brackets.
    """
    kvpstrs = payload.split(",")
    kvp_dict = {}
    for kvpstr in kvpstrs:
        kvp = kvpstr.split("=")
        if len(kvp) == 1:
            kvp_dict[kvp[0].strip()] = ""
        else:
            key, value = kvp
            values = []
            if array_indicator in key:
                for line in all_lines[line_number + 1 :]:
                    if line.startswith("["):
                        break
                    values.append(np.double(line))
                kvp_dict[key.strip()] = values
            else:
                kvp_dict[key.strip()] = value.strip()
    return kvp_dict


def parse_logfile_line(line, line_number, all_lines):
    """
    This function takes a line from a SUNDIALS log file and parses it into a dictionary.
    A log file line has the form:
      [loglvl][rank][scope][label] key1 = value, key2 = value
    Log file payloads can be multiline if they are an array/vector with one value per line.
    I.e.
      [loglvl][rank][scope][label] y(:)
      y_1
      y_2
      ...
    """
    pattern = re.compile(r"\[(\w+)\]\[(rank \d+)\]\[(.*)\]\[(.*)\](.*)")
    matches = pattern.findall(line)
    line_dict = {}
    if matches:
        line_dict["loglvl"] = matches[0][0]
        line_dict["rank"] = matches[0][1]
        line_dict["scope"] = matches[0][2]
        line_dict["label"] = matches[0][3]
        line_dict["payload"] = parse_logfile_payload(
            matches[0][4], line_number, all_lines
        )
    return line_dict


def log_file_to_list(filename, step_scope_txt):
    """
    This function takes a debug log file from a SUNDIALS log file and creates a list where
    each list element represents an integrator step attempt.

    E.g.,
    [
      [
          {
              "loglvl": "DEBUG",
              "rank": "rank 0",
              "scope": "<step_scope_txt>",
              "label": "enter-step-attempt-loop",
              "payload": {"step": "0", "h": "1e-06", "q": "1", "t_n": "0"},
          }, ...
      ], ...
    ]
    """
    with open(filename, "r") as logfile:
        log = []
        lines_for_this_step = None
        all_lines = logfile.readlines()
        for line_number, line in enumerate(all_lines):
            line_dict = parse_logfile_line(line.rstrip(), line_number, all_lines)
            if not line_dict:
                continue
            if (
                line_dict["scope"] == step_scope_txt
                and line_dict["label"] == "enter-step-attempt-loop"
            ):
                if lines_for_this_step is None:
                    lines_for_this_step = [line_dict]
                else:
                    log.append(lines_for_this_step)
                    lines_for_this_step = [line_dict]
            else:
                lines_for_this_step.append(line_dict)
    return log


def cvode_debug_file_to_list(filename):
    """
    This function takes a debug log file from CVODE and creates a list where
    each list entry is a step attempt. See log_file_to_list.
    """
    return log_file_to_list(filename, "CVODE::cvStep")




"""
This is a Python library of utilities SUNDIALS developer may find useful.
Right now it consists of the following modules:

- `logs`: this module has functions for parsing logs produced by `SUNLogger`.

"""



#!/usr/bin/env python3
# ------------------------------------------------------------------------------
# Programmer(s):  Daniel R. Reynolds @ SMU
# ------------------------------------------------------------------------------
# SUNDIALS Copyright Start
# Copyright (c) 2002-2024, Lawrence Livermore National Security
# and Southern Methodist University.
# All rights reserved.
#
# See the top-level LICENSE and NOTICE files for details.
#
# SPDX-License-Identifier: BSD-3-Clause
# SUNDIALS Copyright End
# ------------------------------------------------------------------------------
# matplotlib-based plotting script for the advection_reaction_3D benchmark codes
# ------------------------------------------------------------------------------

# imports
from os.path import exists
import numpy as np
import matplotlib.pyplot as plt

# ------------------------------------------------------------------------------

# utility functions
def parallel_coords(rank):
    if (rank == 0):
        return [0, 0, 0]
    if (rank == 1):
        return [0, 0, 1]
    if (rank == 2):
        return [0, 1, 0]
    if (rank == 3):
        return [0, 1, 1]
    if (rank == 4):
        return [1, 0, 0]
    if (rank == 5):
        return [1, 0, 1]
    if (rank == 6):
        return [1, 1, 0]
    if (rank == 7):
        return [1, 1, 1]

def xslice(u,it,ix):
    return u[it,ix,:,:]

def yslice(u,it,iy):
    return u[it,:,iy,:]

def zslice(u,it,iz):
    return u[it,:,:,iz]

def xproj(u,it):
    return np.average(u[it,:,:,:], axis=0)

def yproj(u,it):
    return np.average(u[it,:,:,:], axis=1)

def zproj(u,it):
    return np.average(u[it,:,:,:], axis=2)

def myplot(axis, X, Y, Z, xlabel='none', ylabel='none'):
    frame = axis.contourf(X, Y, Z)
    plt.colorbar(frame, ax=axis)
    if (xlabel != 'none'):
        axis.set_xlabel(xlabel)
    if (ylabel != 'none'):
        axis.set_ylabel(ylabel)



# read time mesh
times = np.loadtxt("t.000000.txt")
nt = times.size

# read spatial mesh
mesh = np.loadtxt("mesh.txt", dtype=float)
x = mesh[0,:]
y = mesh[1,:]
z = mesh[2,:]
nx = x.size
ny = y.size
nz = z.size

# ensure that the run used exactly 1 or 8 MPI ranks
for i in range(9):
    if (exists("u.00000" + str(i) + ".txt" ) and
        not exists("u.00000" + str(i+1) + ".txt" )):
        nprocs = i+1
if ((nprocs != 1) and (nprocs != 8)):
    print("make_plots.py error: run must have used either 1 or 8 MPI ranks")
    exit()

# load data for run
if (nprocs == 1):
    u = np.zeros((nt,nx,ny,nz), dtype=float)
    v = np.zeros((nt,nx,ny,nz), dtype=float)
    w = np.zeros((nt,nx,ny,nz), dtype=float)
    udata = np.loadtxt("u.000000.txt")
    vdata = np.loadtxt("v.000000.txt")
    wdata = np.loadtxt("w.000000.txt")
    if (nt != udata.shape[0]):
        print("make_plots.py error: mesh and data have incompatible sizes")
        exit()
    if (nx*ny*nz != udata.shape[1]):
        print("make_plots.py error: mesh and data have incompatible sizes")
        exit()
    for it in range(nt):
        u[it,:,:,:] = np.reshape(udata[it,:], (nx,ny,nz), order='C')
        v[it,:,:,:] = np.reshape(vdata[it,:], (nx,ny,nz), order='C')
        w[it,:,:,:] = np.reshape(wdata[it,:], (nx,ny,nz), order='C')
else:
    u = np.zeros((nt,nx,ny,nz), dtype=float)
    v = np.zeros((nt,nx,ny,nz), dtype=float)
    w = np.zeros((nt,nx,ny,nz), dtype=float)
    nxl = nx//2
    nyl = ny//2
    nzl = nz//2
    for ip in range(8):
        udata = np.loadtxt("u.00000" + str(ip) + ".txt")
        vdata = np.loadtxt("v.00000" + str(ip) + ".txt")
        wdata = np.loadtxt("w.00000" + str(ip) + ".txt")
        if (nt != udata.shape[0]):
            print("make_plots.py error: mesh and data have incompatible sizes")
            exit()
        if (nxl*nyl*nzl != udata.shape[1]):
            print("make_plots.py error: mesh and data have incompatible sizes")
            exit()
        coords = parallel_coords(ip)
        ilo = coords[0]*nxl
        ihi = (coords[0]+1)*nxl
        jlo = coords[1]*nyl
        jhi = (coords[1]+1)*nyl
        klo = coords[2]*nzl
        khi = (coords[2]+1)*nzl
        for it in range(nt):
            u[it,ilo:ihi,jlo:jhi,klo:khi] = np.reshape(udata[it,:], (nxl,nyl,nzl), order='C')
            v[it,ilo:ihi,jlo:jhi,klo:khi] = np.reshape(vdata[it,:], (nxl,nyl,nzl), order='C')
            w[it,ilo:ihi,jlo:jhi,klo:khi] = np.reshape(wdata[it,:], (nxl,nyl,nzl), order='C')


# set meshgrid objects
xy0,xy1 = np.meshgrid(x, y)
yz0,yz1 = np.meshgrid(y, z)
xz0,xz1 = np.meshgrid(x, z)

# generate plots
sliceidx = 25
tslice = [0, 5, 10]
figsize = (9,7)

#    xy slices at various times
plt.figure(1)
fig, ((ax1,ax2,ax3), (ax4,ax5,ax6), (ax7,ax8,ax9)) = plt.subplots(3, 3, sharex=True, sharey=True, figsize=figsize)
myplot(ax1, xy0, xy1, zslice(u,tslice[0],sliceidx), ylabel = 'u')
myplot(ax2, xy0, xy1, zslice(u,tslice[1],sliceidx))
myplot(ax3, xy0, xy1, zslice(u,tslice[2],sliceidx))
myplot(ax4, xy0, xy1, zslice(v,tslice[0],sliceidx), ylabel = 'v')
myplot(ax5, xy0, xy1, zslice(v,tslice[1],sliceidx))
myplot(ax6, xy0, xy1, zslice(v,tslice[2],sliceidx))
myplot(ax7, xy0, xy1, zslice(w,tslice[0],sliceidx), ylabel = 'w', xlabel = 't = ' + str(times[0]))
myplot(ax8, xy0, xy1, zslice(w,tslice[1],sliceidx), xlabel = 't = ' + str(times[1]))
myplot(ax9, xy0, xy1, zslice(w,tslice[2],sliceidx), xlabel = 't = ' + str(times[2]))
plt.savefig('xy-slices.png')

#    yz slices at various times
plt.figure(2)
fig, ((ax1,ax2,ax3), (ax4,ax5,ax6), (ax7,ax8,ax9)) = plt.subplots(3, 3, sharex=True, sharey=True, figsize=figsize)
myplot(ax1, yz0, yz1, xslice(u,tslice[0],sliceidx), ylabel = 'u')
myplot(ax2, yz0, yz1, xslice(u,tslice[1],sliceidx))
myplot(ax3, yz0, yz1, xslice(u,tslice[2],sliceidx))
myplot(ax4, yz0, yz1, xslice(v,tslice[0],sliceidx), ylabel = 'v')
myplot(ax5, yz0, yz1, xslice(v,tslice[1],sliceidx))
myplot(ax6, yz0, yz1, xslice(v,tslice[2],sliceidx))
myplot(ax7, yz0, yz1, xslice(w,tslice[0],sliceidx), ylabel = 'w', xlabel = 't = ' + str(times[0]))
myplot(ax8, yz0, yz1, xslice(w,tslice[1],sliceidx), xlabel = 't = ' + str(times[1]))
myplot(ax9, yz0, yz1, xslice(w,tslice[2],sliceidx), xlabel = 't = ' + str(times[2]))
plt.savefig('yz-slices.png')

#    xz slices at various times
plt.figure(3)
fig, ((ax1,ax2,ax3), (ax4,ax5,ax6), (ax7,ax8,ax9)) = plt.subplots(3, 3, sharex=True, sharey=True, figsize=figsize)
myplot(ax1, xz0, xz1, yslice(u,tslice[0],sliceidx), ylabel ='u')
myplot(ax2, xz0, xz1, yslice(u,tslice[1],sliceidx))
myplot(ax3, xz0, xz1, yslice(u,tslice[2],sliceidx))
myplot(ax4, xz0, xz1, yslice(v,tslice[0],sliceidx), ylabel = 'v')
myplot(ax5, xz0, xz1, yslice(v,tslice[1],sliceidx))
myplot(ax6, xz0, xz1, yslice(v,tslice[2],sliceidx))
myplot(ax7, xz0, xz1, yslice(w,tslice[0],sliceidx), ylabel= 'w', xlabel = 't = ' + str(times[0]))
myplot(ax8, xz0, xz1, yslice(w,tslice[1],sliceidx), xlabel ='t = ' + str(times[1]))
myplot(ax9, xz0, xz1, yslice(w,tslice[2],sliceidx), xlabel = 't = ' + str(times[2]))
plt.savefig('xz-slices.png')

#    xy projection at various times
plt.figure(4)
fig, ((ax1,ax2,ax3), (ax4,ax5,ax6), (ax7,ax8,ax9)) = plt.subplots(3, 3, sharex=True, sharey=True, figsize=figsize)
myplot(ax1, xy0, xy1, zproj(u,tslice[0]), ylabel = 'u')
myplot(ax2, xy0, xy1, zproj(u,tslice[1]))
myplot(ax3, xy0, xy1, zproj(u,tslice[2]))
myplot(ax4, xy0, xy1, zproj(v,tslice[0]), ylabel = 'v')
myplot(ax5, xy0, xy1, zproj(v,tslice[1]))
myplot(ax6, xy0, xy1, zproj(v,tslice[2]))
myplot(ax7, xy0, xy1, zproj(w,tslice[0]), ylabel = 'w', xlabel = 't = ' + str(times[0]))
myplot(ax8, xy0, xy1, zproj(w,tslice[1]), xlabel = 't = ' + str(times[1]))
myplot(ax9, xy0, xy1, zproj(w,tslice[2]), xlabel = 't = ' + str(times[2]))
plt.savefig('xy-projections.png')

#    yz projection at various times
fig = plt.figure(5)
fig, ((ax1,ax2,ax3), (ax4,ax5,ax6), (ax7,ax8,ax9)) = plt.subplots(3, 3, sharex=True, sharey=True, figsize=figsize)
myplot(ax1, yz0, yz1, xproj(u,tslice[0]), ylabel = 'u')
myplot(ax2, yz0, yz1, xproj(u,tslice[1]))
myplot(ax3, yz0, yz1, xproj(u,tslice[2]))
myplot(ax4, yz0, yz1, xproj(v,tslice[0]), ylabel = 'v')
myplot(ax5, yz0, yz1, xproj(v,tslice[1]))
myplot(ax6, yz0, yz1, xproj(v,tslice[2]))
myplot(ax7, yz0, yz1, xproj(w,tslice[0]), ylabel = 'w', xlabel = 't = ' + str(times[0]))
myplot(ax8, yz0, yz1, xproj(w,tslice[1]), xlabel = 't = ' + str(times[1]))
myplot(ax9, yz0, yz1, xproj(w,tslice[2]), xlabel = 't = ' + str(times[2]))
plt.savefig('yz-projections.png')

#    xz projection at various times
fig = plt.figure(6)
fig, ((ax1,ax2,ax3), (ax4,ax5,ax6), (ax7,ax8,ax9)) = plt.subplots(3, 3, sharex=True, sharey=True, figsize=figsize)
myplot(ax1, xz0, xz1, yproj(u,tslice[0]), ylabel = 'u')
myplot(ax2, xz0, xz1, yproj(u,tslice[1]))
myplot(ax3, xz0, xz1, yproj(u,tslice[2]))
myplot(ax4, xz0, xz1, yproj(v,tslice[0]), ylabel = 'v')
myplot(ax5, xz0, xz1, yproj(v,tslice[1]))
myplot(ax6, xz0, xz1, yproj(v,tslice[2]))
myplot(ax7, xz0, xz1, yproj(w,tslice[0]), ylabel = 'w', xlabel = 't = ' + str(times[0]))
myplot(ax8, xz0, xz1, yproj(w,tslice[1]), xlabel = 't = ' + str(times[1]))
myplot(ax9, xz0, xz1, yproj(w,tslice[2]), xlabel = 't = ' + str(times[2]))
plt.savefig('xz-projections.png')

#plt.show()
plt.close()

##### end of script #####



# -----------------------------------------------------------------------------
# SUNDIALS Copyright Start
# Copyright (c) 2002-2024, Lawrence Livermore National Security
# and Southern Methodist University.
# All rights reserved.
#
# See the top-level LICENSE and NOTICE files for details.
#
# SPDX-License-Identifier: BSD-3-Clause
# SUNDIALS Copyright End
# -----------------------------------------------------------------------------
# SUNDIALS Sphinx extension
# -----------------------------------------------------------------------------

from sphinx.application import Sphinx

def setup(app: Sphinx):
    # Create new object type for CMake options
    app.add_object_type('cmakeoption', 'cmakeop', 'single: CMake options; %s')
    # Create new configuration value set in conf.py
    app.add_config_value('package_name', '', 'env', types=[str])



#!/usr/bin/env python3
# -----------------------------------------------------------------------------
# Programmer(s): Yu Pan @ LLNL
# -----------------------------------------------------------------------------
# SUNDIALS Copyright Start
# Copyright (c) 2002-2024, Lawrence Livermore National Security
# and Southern Methodist University.
# All rights reserved.
#
# See the top-level LICENSE and NOTICE files for details.
#
# SPDX-License-Identifier: BSD-3-Clause
# SUNDIALS Copyright End
# -----------------------------------------------------------------------------

import os
import subprocess
import sys
import re
import glob
import argparse
import multiprocessing as mp
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

import hatchet as ht
import thicket as tt

def main():
    parser = argparse.ArgumentParser(description='Compare Sundials performance results against previous results')

    parser.add_argument('--release', dest='release', action='store_true', help='indicate if the current run to process is a release')

    parser.add_argument('--calidir', dest='caliDir', type=str, help='path to directory containing caliper files', default="/usr/workspace/sundials/califiles")

    parser.add_argument('--releasedir', dest='releaseDir', type=str, help='path to directory containing release caliper files', default="/usr/workspace/sundials/califiles/Release")

    parser.add_argument('--outpath', dest='outPath', type=str, help='path to directory to write results to', default="/dev/null")

    parser.add_argument('--threshold', dest="threshold", type=float, help='the percentage threshold in performance difference that indicates a regression', default=2.0)

    args = parser.parse_args()

    release = args.release
    releaseDir = args.releaseDir
    caliDir = args.caliDir
    outPath = args.outPath
    threshold = args.threshold

    # Get the latest test run
    runDirs = glob.glob("%s/Testing/*" % caliDir, recursive = True)
    runDirs.sort(key=os.path.getmtime, reverse=True)
    runDir = runDirs[0]

    runFile = glob.glob(runDir)[0]
    th_temp = tt.Thicket.from_caliperreader(runFile)
    cluster = th_temp.metadata['cluster']
    # get machine from the file
    if release:
        # Compare against the last release
        versionDirs = glob.glob("%s/%s/*" % (releaseDir, cluster))
        versionDirs.sort(key=os.path.getmtime, reverse=True)
        versionDir = versionDirs[1]
    else:
        # Compare against the release the run is a part of
        version = th_temp.metadata['sundials_version'].values[0]
        versionDir = "%s/%s/%s" % (releaseDir, cluster, version)

    # Gather files to process
    runFiles = glob.glob("%s/*.cali" % (runDir))

    if not os.path.exists(outPath):
        os.makedirs(outPath)
    outFile = open("%s/output.out" % outPath, 'w')

    # Compare test results against past runs. If a test performs below a threshold, output test name to outFile.
    with mp.Pool() as pool:
        for res in pool.starmap(compare_against_release, [(versionDir, i, threshold) for i in runFiles]):
            if res:
                outFile.write(res + "\n")
    outFile.close()

    outFile = open("%s/example_output.out" % outPath, 'r')
    try:
        outLines = outFile.readlines()
    finally:
        outFile.close()

    if (len(outLines) == 0):
        return -1
    return 0


def compare_against_release(releaseDir, file, threshold):
    th = tt.Thicket.from_caliperreader(file)

    testName = th.metadata['env.TEST_NAME'].values[0]

    # Gather release run
    releaseFile = glob.glob("%s/Testing/*/%s.*.cali" % (releaseDir, testName), recursive=True)
    th_release = tt.Thicket.from_caliperreader(releaseFile)

    metrics = ['Max time/rank']
    tt.mean(th_release, columns=metrics)
    tt.mean(th, columns=metrics)

    ratio = th.statsframe.dataframe['Max time/rank_mean'] / th_release.statsframe.dataframe['Max time/rank_mean']
    print(ratio[0])
    tolerance = threshold/100
    if 1 - ratio[0] < tolerance:
        return testName

if __name__ == "__main__":
    main()
